/*
 * Copyright (C) 2020 Linux Studio Plugins Project <https://lsp-plug.in/>
 *           (C) 2020 Vladimir Sadovnikov <sadko4u@gmail.com>
 *
 * This file is part of lsp-plugins-shared
 * Created on: 25 нояб. 2020 г.
 *
 * lsp-plugins-shared is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * any later version.
 *
 * lsp-plugins-shared is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with lsp-plugins-shared. If not, see <https://www.gnu.org/licenses/>.
 */

#include <lsp-plug.in/shared/meta/developers.h>

namespace lsp
{
    namespace developers
    {
        const meta::person_t v_sadovnikov =
        {
            "v_sadovnikov",
            "SadKo",
            "Vladimir Sadovnikov",
            "sadko4u@gmail.com",
            "http://lsp-plug.in/"
        };

        const meta::person_t s_tronci =
        {
            "s_tronci",
            "CrocoDuck",
            "Stefano Tronci",
            "crocoduck.oducks@gmail.com",
            "https://thecrocoduckspond.wordpress.com/"
        };

    }
}



