/*
 * Copyright (C) 2026 Linux Studio Plugins Project <https://lsp-plug.in/>
 *           (C) 2026 Vladimir Sadovnikov <sadko4u@gmail.com>
 *
 * This file is part of lsp-plugin-fw
 * Created on: 15 янв. 2026 г.
 *
 * lsp-plugin-fw is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * any later version.
 *
 * lsp-plugin-fw is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with lsp-plugin-fw. If not, see <https://www.gnu.org/licenses/>.
 */

#ifndef LSP_PLUG_IN_PLUG_FW_UTIL_LAUNCHER_WINDOW_H_
#define LSP_PLUG_IN_PLUG_FW_UTIL_LAUNCHER_WINDOW_H_

#include <lsp-plug.in/plug-fw/version.h>

#include <lsp-plug.in/plug-fw/ctl.h>

namespace lsp
{
    namespace launcher
    {
        class Window: public ctl::Window
        {
            public:
                explicit Window(ui::IWrapper *src, tk::Window *window);
                Window(const Window &) = delete;
                Window(Window &&) = delete;
                virtual ~Window() override;
                Window & operator = (const Window &) = delete;
                Window & operator = (Window &&) = delete;

                virtual status_t        init() override;
                virtual void            destroy() override;

            public:
                status_t                post_init();
        };


    } /* namespace launcher */
} /* namespace lsp */



#endif /* LSP_PLUG_IN_PLUG_FW_UTIL_LAUNCHER_WINDOW_H_ */
