/*
 * Copyright (C) 2025 Linux Studio Plugins Project <https://lsp-plug.in/>
 *           (C) 2025 Vladimir Sadovnikov <sadko4u@gmail.com>
 *
 * This file is part of lsp-ws-lib
 * Created on: 27 Dec 2025
 *
 * lsp-ws-lib is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * any later version.
 *
 * lsp-ws-lib is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with lsp-ws-lib. If not, see <https://www.gnu.org/licenses/>.
 */

#ifndef PRIVATE_COCOA_DEFS_H_
#define PRIVATE_COCOA_DEFS_H_

#include <private/cocoa/fwd_defs.h>

#import <Cocoa/Cocoa.h>

namespace lsp
{
    namespace ws
    {
        namespace cocoa
        {
            struct nsevent_t
            {
                NSEvent * const event;
            };
        
            struct nswindow_t
            {
                NSWindow * const window;
            };
        
        } /* namespace cocoa */
    } /* namespace ws */
} /* namespace lsp */

#endif /* PRIVATE_COCOA_DEFS_H_ */
